//UnitParameters.java

import swarm.Globals;
import swarm.defobj.Zone;
import swarm.objectbase.SwarmObjectImpl;
import java.io.*;
import java.util.*;

/**
 * This class reads the parameters (costs and prodocution phases) for each unit
 *
 * @author Dario Landini and Nicola Ormezzano; IO improved by Michele Sonnessa
 * @author deeply modified by Pietro Terna to deal with simple or complex
 * units
 */

public class UnitParameters extends SwarmObjectImpl 
{

    /** the vector of the unit numbers */
    public int[] unitNumber;
    /** the vector of simple unit production phase (simple=with a unique 
	activity) */
    public int[] simpleUnitProductionPhase; 
    /** the vector of the fixed costs (per day or time step) of each unit */
    public float[] fixedCosts;
    /** the vector of the variable costs (per day or time step) of each unit */
    public float[] variableCosts;
    /** the vector of local decisione about the use of warehouses; if global
     *  decision is negative, the local one is ininfluent; if the global
     *  decisione, local choice is relevant; to use warehouses we must have */ 
    /** the number of the units (the size of the vectors) */
    public int[] useWarehouseLocalChoice;
    public int totalUnitNumber;
    /** check if we are using an Excel like (generated by Office or OpenOffice)
     *  file containing special unit data */
    public boolean worksheetUnitFileOpen=false;
    /** the worksheet containing special unit data, if any */
    ExcelReader specialUnitWorksheet = null;
    int i, ii;

    /** the constructor of UnitParameters
     */
    public UnitParameters(Zone aZone, int un) 
    {

	// call the constructor for the UnitParameters parent class
	super(aZone);
	totalUnitNumber = un;
    }

    /** read the unit number, production phase, fixed cost and variable cost
     *  parameters from the unitBasicData.txt file*/
    public void readUnitData(){
	try{
	    unitNumber = new int[totalUnitNumber];
	    simpleUnitProductionPhase = new int[totalUnitNumber];
	    fixedCosts = new float[totalUnitNumber];
	    variableCosts = new float[totalUnitNumber];
	    useWarehouseLocalChoice = new int[totalUnitNumber];

	    // read the file
	    BufferedReader fileUnitBasicData = null;
	    StringTokenizer t;
	    fileUnitBasicData = new 
		BufferedReader(new FileReader("unitData/unitBasicData.txt"));

	    // reading a control line (titles)
	    String line = " ", lineTest = 
      "unit_#__useWarehouse____prod.phase_#____fixed_costs_____variable_costs";

	    line=fileUnitBasicData.readLine();
	    if (! line.equals(lineTest))
		{
		    System.out.println(
		    "First line in unitData/unitBasicData.txt must contain:");
		    System.out.println(lineTest);
		    System.exit(1);
		}
	    line=" ";

	    // starting with a void line; the actual
	    // reading is done in MyReader, dealing also
	    // with the missing data case
	    t = new StringTokenizer(line, " ");

	    for(i = 0; i < totalUnitNumber; i++)
		{
		    t=MyReader.check("unitData/unitBasicData.txt", 
				     fileUnitBasicData, line, t);
		    unitNumber[i]    = Integer.parseInt(t.nextToken());
		    useWarehouseLocalChoice[i] = Integer.
			parseInt(t.nextToken());
		    simpleUnitProductionPhase[i] 
			             = Integer.parseInt(t.nextToken());
		    fixedCosts[i]    = Float.parseFloat(t.nextToken());
		    variableCosts[i] = Float.parseFloat(t.nextToken());
		}

	    fileUnitBasicData.close();
	}
	catch (FileNotFoundException fnfe)
	    {
		System.out.println(fnfe);
		System.exit(1);
	    } 
	catch (IOException e)
	    {
		System.out.println("IOException:" + e.toString());
	    }
    }

    /** setting the parameters of units */
    public void setParametersTo(int i, Unit aUnit){
	String fileName; 
	float fixedCosts, variableCosts;
	int unitProductionPhase, nOfPhasesToDealWith, phaseInventories;

	aUnit.setUnitNumber(getUnitNumber(i));
	aUnit.setPhaseInventories(getUseWarehouseLocalChoice(i));
	aUnit.setProductionPhase(getSimpleUnitProductionPhase(i)); 
	aUnit.setFixedAndVariableCosts(getFixedCosts(i), getVariableCosts(i));


	// complex unit case, with a sheet in a workshhet file related to
	// the each unit
	if(getSimpleUnitProductionPhase(i) == 0) // i.e. no unique phase
	    {
		// open (once) the Excel file containing special unit data */
		if(! worksheetUnitFileOpen) 
		    specialUnitWorksheet = 
			new ExcelReader("unitData/units.xls");
		worksheetUnitFileOpen = true;

		specialUnitWorksheet.selectSheet(String.valueOf(i));

		if(StartVEFrame.verbose)
		    System.out.println("Unit " + getUnitNumber(i) 
				       + " is a special one.");

		// the unit is able to deal with nOfPhasesToDealWith
		// different production phases
		nOfPhasesToDealWith=specialUnitWorksheet.getIntValue();
		if(StartVEFrame.verbose)
     	           System.out.println("special unit " + 
                   getUnitNumber(i) + " nOfPhasesToDealWith " +
		   nOfPhasesToDealWith);

		aUnit.setNOfPhasesToDealWith(nOfPhasesToDealWith);

		for (ii=1; ii<=nOfPhasesToDealWith; ii++)
		    {
			// production phases (one or more)
			unitProductionPhase=specialUnitWorksheet.getIntValue();
			aUnit.setProductionPhase(unitProductionPhase);

			// costs for each phase
 
			fixedCosts = (float) 
			    specialUnitWorksheet.getDblValue();
			variableCosts = (float) 
			    specialUnitWorksheet.getDblValue();
			    if(StartVEFrame.verbose)
			       System.out.println("special unit " + 
			       getUnitNumber(i) + " fixed costs " + 
			       fixedCosts + " variable costs " + 
						  variableCosts);
			    aUnit.setFixedAndVariableCosts(fixedCosts, 
							   variableCosts);
			    phaseInventories = 
				specialUnitWorksheet.getIntValue();
			    aUnit.setPhaseInventories(phaseInventories);
		    }
	    }

   
    }
    
    /** returning the unit number for a specific unit */
    public int getUnitNumber(int i){
	return unitNumber[i-1];
    }
 
    /** returning the production phase for a specific simple unit */
    public int getSimpleUnitProductionPhase(int i){
	return simpleUnitProductionPhase[i-1];
    }

    /** returning the fixed cost for a specific unit */
    public float getFixedCosts(int i){
	return fixedCosts[i-1];
    }
    
    /** returning the variable cost for a specific unit */
    public float getVariableCosts(int i){
	return variableCosts[i-1];
    }

    /** returning the local use of the warehouse for a specific unit */
    public int getUseWarehouseLocalChoice (int i){
	return useWarehouseLocalChoice[i-1];
    }
    
}
