// VEFrameModelSwarm.java

import swarm.Globals;
import swarm.Selector;
import swarm.defobj.Zone;
import swarm.defobj.SymbolImpl;

import swarm.activity.Activity;
import swarm.activity.ActionGroup;
import swarm.activity.ActionGroupImpl;
import swarm.activity.Schedule;
import swarm.activity.ScheduleImpl;

import swarm.objectbase.Swarm;
import swarm.objectbase.SwarmImpl;
import swarm.objectbase.VarProbe;
import swarm.objectbase.MessageProbe;
import swarm.objectbase.EmptyProbeMapImpl;

import swarm.collections.ListImpl;

import java.io.*;
import java.util.*;

/**
 * The Model contains the Units and all the related tools, like the
 * warehouses which are necessary to deal with the inventories or the
 * newses
 *
 * @author Pietro Terna, with contributions of Marco Remondino, Dario Landini, 
 * Nicola Ormezzano and Paolo Pelligra
 */
public class VEFrameModelSwarm extends SwarmImpl
{
    // simulation parameters

    /** the number of ticks of the clock in a day; at each tick a unit
     * of each step in recipes is done 
     */ 
    public int ticksInADay;
    /** the total number of operating Units we are using*/
    public int totalUnitNumber;
    /** the max number of steps to be done to complete an order*/
    public int maxStepNumber;
    /** the max length of a step in a recipe (in units of time or ticks) */
    public int maxStepLength;
    /** the max inventory quantity in a warehouse */
    public int maxInWarehouses;
    /** the min inventory quantity in a warehouse */
    public int minInWarehouses;
    /** choosing if we use or not warehouses (and inventories) */
    public boolean useWarehouses;
    /** choosing if we use or not newses  */
    public boolean useNewses;
    /** how deep information are propagated (how many steps we examine after
     * the current production phase)*/
    public int infDeepness;
    /** the estimate of the revenue of the enterprise per each recipe step 
     * done in a semi manufactured order */
    public float revenuePerEachRecipeStep;
    /** the yearly rate used to evaluate the financial costs of inventories */
     public  float inventoryFinancialRate;

    /** the number of newses necessary to decide the production of 
     *  inventories */
    public int nOfNewsesToProduce;
    /** the number of newses to be cleared after the decision of producing
     *  to increase inventories (a simple choice is to have it equal to 
     *  nOfNewsesToProduce) */
    public int nOfNewsesToBeCleared;
    /** the number of orders for which newses are propagated by a unit */
    public int nOfOrdersInNewses;
    /** the inventory evaluation criterion */
    public  int inventoryEvaluationCriterion;
    /** the total production time used by done orders */
    public int totalProductionTime;
    /** the total length of the recipes of the done orders */
    public int totalRecipeLength;

    /** the order generator, to be replaced in the future by more
      sophisticated structures (a swarm of enterprises etc.) */
    public OrderGenerator orderGenerator;

    /** ActionGroup for holding an ordered sequence of action */
    public ActionGroup modelActions1, modelActions2, modelActions3;

    /** the Schedule operating in the Model*/
    public Schedule modelSchedule;

    /** the inventoryRuleMaster managing inventories */
    public InventoryRuleMaster inventoryRuleMaster;

    /** the informationRuleMaster managing newses */
    public InformationRuleMaster informationRuleMaster;
    /** the input class for unit parameters*/
    public UnitParameters unitParameters;

    /** the list of the operating units */
    public ListImpl unitList;
    /** the list of the warehouses */
    public ListImpl warehouseList;
    /** the whole list of the orders (not yet finished) */
    public ListImpl orderList;


    public VEFrameModelSwarm (Zone aZone) {
	super (aZone);

	// Now fill in various simulation parameters with default values.
	ticksInADay = 1;
	totalUnitNumber = 10;
	maxStepNumber=5;
	maxStepLength=1;
	useWarehouses=true;
	maxInWarehouses=3;
	minInWarehouses=3;
	useNewses=true;
	infDeepness=3;
	inventoryFinancialRate= (float) 0.1;
	inventoryEvaluationCriterion=1;
	revenuePerEachRecipeStep=2;
	nOfNewsesToProduce=1;
	nOfNewsesToBeCleared=100000;
	nOfOrdersInNewses=1;

	// Build a customized probe map using a `local' subclass
	// (a special kind of Java `inner class') of the
	// EmptyProbeMapImpl class.
	// This building operation produces a separated class file, with
	// name  VEFrameModelSwarm$1$VEFrameModelProbeMap.class
	class VEFrameModelProbeMap extends EmptyProbeMapImpl {
	    public VarProbe probeVariable (String name) {
		return
		    Globals.env.probeLibrary.getProbeForVariable$inClass
		    (name, VEFrameModelSwarm.this.getClass ());
	    }
	    public MessageProbe probeMessage (String name) {
		return
		    Globals.env.probeLibrary.getProbeForMessage$inClass
		    (name, VEFrameModelSwarm.this.getClass ());
	    }
	    public void addVar (String name) {
		addProbe (probeVariable (name));
	    }
	    public void addMessage (String name) {
		addProbe (probeMessage (name));
	    }
	    public VEFrameModelProbeMap (Zone _aZone, Class aClass) {
		super (_aZone, aClass);
		addVar ("ticksInADay");
		addVar ("totalUnitNumber");
		addVar ("maxStepNumber");
		addVar ("maxStepLength");
		addVar ("useWarehouses");
		addVar ("useNewses");
		addVar ("maxInWarehouses");
		addVar ("minInWarehouses");
		addVar ("infDeepness");
		addVar ("inventoryFinancialRate");
		addVar ("inventoryEvaluationCriterion");
		addVar ("revenuePerEachRecipeStep");
		addVar ("nOfNewsesToProduce");
		addVar ("nOfNewsesToBeCleared");
		addVar ("nOfOrdersInNewses");
	    }
	}

	// Install our custom probeMap class directly into the
	// probeLibrary
	Globals.env.probeLibrary.setProbeMap$For
	    (new VEFrameModelProbeMap (aZone, getClass ()), getClass ());
    }
    /** Build the model objects. */
    public Object buildObjects ()
    {
	int i;

	// allow our parent class to build anything
	super.buildObjects();

        // create the class to read acounting parameters
        unitParameters = new 
	    UnitParameters(getZone(), totalUnitNumber);
        // read fixed and variable costs
        unitParameters.readUnitData();

	// check maxInWarehouses value and if 0 change it to the max
	// integer value
	if (maxInWarehouses==0)
	    {maxInWarehouses=Integer.MAX_VALUE;
	    // no verbose control here
	    System.out.println(
      "Warning, maxInWarehouses has been changed from 0 to Integer.MAX_VALUE");
	    }

	// the list of the production units
	unitList = new ListImpl (getZone());
	// the list of the warehouses
	warehouseList = new ListImpl (getZone());
        // the whole list of the orders existing at each time step
        orderList = new ListImpl(getZone());

	// the orderGenerator
	orderGenerator = new OrderGenerator
	    (getZone(), maxStepNumber, maxStepLength,unitList, orderList, 
	     this);

	// the inventoryRuleMaster managing the inventories
	inventoryRuleMaster = new 
	    InventoryRuleMaster(getZone(),
				maxInWarehouses, minInWarehouses,
				nOfNewsesToProduce, nOfNewsesToBeCleared);

        // the informationRuleMaster managing the news propagation
        if(useWarehouses && useNewses)
        informationRuleMaster = new InformationRuleMaster(getZone(),
							  totalUnitNumber);

	// create the units and their warehouses, if any
	for (i = 1; i <= totalUnitNumber; i++)
	    {
		Warehouse aWarehouse=null; // null to avoid errors when unused
		News aNews=null;           //  "   "   "     "      "    "
		Unit aUnit;

		// using warehouses and not newses
		if(useWarehouses && ! useNewses)
		{
		    // using the unit number as id also for its warehouse
		    aWarehouse =  new Warehouse (getZone(), unitParameters.
						 getUnitNumber(i));

		    aUnit  =  new Unit (getZone(), unitList, this,
					inventoryRuleMaster,  aWarehouse,
					orderList);
		    unitParameters.setParametersTo(i, aUnit);
		}
		// using both warehoueses and newses
		else if(useWarehouses && useNewses)
		    {
			// using the unit number as id also for its warehouse
			aWarehouse =  new Warehouse (getZone(), unitParameters.
						     getUnitNumber(i));
			aUnit  =  new Unit(getZone(), unitList, this,
					   inventoryRuleMaster,  aWarehouse,
					   informationRuleMaster, infDeepness, 
				           nOfOrdersInNewses,
				           orderList);
			unitParameters.setParametersTo(i, aUnit);
		    }
		// no warehouses and no newses
		else 
		    {
			aUnit = new Unit  (getZone(), unitList, this,
					   orderList);
			unitParameters.setParametersTo(i, aUnit);
		    }
		// add the unit and the warehouse to the lists
		unitList.addLast (aUnit);
		if(useWarehouses) warehouseList.addLast (aWarehouse);

	    }

	if(! useWarehouses)if(StartVEFrame.verbose)
	    System.out.println("No warehouses generated");

	return this;
    }

    /**
     * Here is where the model schedule is built, the data structures
     * that define the simulation of time in the model. The core is an
     * actionGroup that has a list of actions. That's then put in a
     * Schedule.  */
    public Object buildActions ()
    {
	int i;
	super.buildActions();

	modelActions1 = new ActionGroupImpl (getZone ());
	modelActions2 = new ActionGroupImpl (getZone ());
	modelActions3 = new ActionGroupImpl (getZone ());

	/* zeroing daily accounting variables */
	modelActions1.createActionForEach$message(unitList,
			         SwarmUtils.getSelector("Unit","unitStep0"));
	/* producing and using inventories */
	modelActions2.createActionForEach$message(unitList,
			         SwarmUtils.getSelector("Unit","unitStep1"));
	/* a new order; this step is placed here, after the production step,
	 * to align the diffusion of the order forms (orders under execution 
	 * - next step - or new ones - this step -) */
	modelActions2.createActionTo$message(orderGenerator,
		       	 SwarmUtils.getSelector(orderGenerator,
		       		       	"createRandomOrderWithNSteps"));
	/* diffusing products (order forms) and making accounting per tick */
	modelActions2.createActionForEach$message(unitList,
				 SwarmUtils.getSelector("Unit","unitStep2"));
	/* propagating newses */
	modelActions2.createActionForEach$message(unitList,
                                 SwarmUtils.getSelector("Unit","unitStep3"));
	/* making daily accounting */
	modelActions3.createActionForEach$message(unitList,
				 SwarmUtils.getSelector("Unit","unitStep4"));

	// Then we create a schedule that executes the
	// modelActions. modelActions is an ActionGroup, by itself it
	// has no notion of time. In order to have it executed in
	// time, we create a Schedule that says to use the
	// modelActions ActionGroup at particular times
	modelSchedule = new ScheduleImpl (getZone (), ticksInADay);

	modelSchedule.at$createAction (0, modelActions1);

	for (i=0;i<ticksInADay;i++)
	modelSchedule.at$createAction (i, modelActions2);

	modelSchedule.at$createAction (ticksInADay-1, modelActions3);
	
	return this;
    }

    /**
     * Now set up the model's activation. swarmContext indicates where
     * we're being started in - typically, this model is run as a
     * subswarm of an observer swarm. */
    public Activity activateIn (Swarm swarmContext) {
	// First, activate ourselves via the superclass
	// activateIn: method.  Just pass along the context: the
	// activity library does the right thing.
	super.activateIn (swarmContext);

	// Now activate our own schedule.
	modelSchedule.activateIn (this);

	// Finally, return our activity.
	return getActivity ();
  }
    
    /** the method returns the list of the units */
    public Object getUnitList(){
	return unitList;
    }

    /** the method returns the list of the warehouses */
    public Object getWarehouseList (){
	return warehouseList;
    }

    /** the method returns the list of the orders */
    public Object getOrderList(){
        return orderList;
    }

    /** this get method returns true if warehouses are used, false otherwise */
    public boolean getUseWarehouses() {
	return useWarehouses;
    }

    /** record total production time and total recipe length */
    public void setProductionTimeAndRecipeLength(int pt, int rl) {
	totalProductionTime += pt;
	totalRecipeLength   += rl;
	if(StartVEFrame.verbose)
	    System.out.println("totalProductionTime " + totalProductionTime +
			       " totalRecipeLength " + totalRecipeLength +
			       " ratio " + ((float)
					    totalProductionTime)/totalRecipeLength);
	return;
    }

    /** report the ratio totalProductionTime/totalRecipeLength */
    public float getTimeLengthRatio() {
	if(totalRecipeLength != 0)
	    return ((float) totalProductionTime)/totalRecipeLength;
	else return (float) 1; // 1 is a good choice, being 0 totalRecipeLength
	                       // in the first simulation days and being 1
	                       // the actual ratio of the first
	                       // concluded orders
    }

    /** the financial rate */
    public float getInventoryFinancialRate(){
        return inventoryFinancialRate;
    }

    /** the evaluation criterion for inventories */
    public float getInventoryEvaluationCriterion(){
        return inventoryEvaluationCriterion;
    }

   /** the revenue for each recipe step */
    public float getRevenuePerEachRecipeStep(){
        return revenuePerEachRecipeStep;
    }

    /** the number of ticks in a day */
    public int getTicksInADay(){
	return ticksInADay;
    }

    /** the benefit of the enterprise */
    public float getBenefit() {
	float benefit;
        float totalRevenues=0;
        float totalCosts=0;
        float totalSemimanufacturedRevenues=0;
        float dailyStoredComponentRevenues=0;
        float totalInventoryFinCosts=0;
        Unit aUnit;
        Order anOrder;
	int i;

        for(i=0; i<unitList.getCount(); i++){
            aUnit=(Unit)unitList.atOffset(i);
            totalRevenues += aUnit.getTotalSaleRevenues();
            totalCosts += aUnit.getTotalCosts();
            if(useWarehouses){
		dailyStoredComponentRevenues += 
		    aUnit.getDailyStoredComponentValue();
		totalInventoryFinCosts += 
		    aUnit.getTotalInventoryFinancialCosts();
            }
        }

        for(i=0; i<orderList.getCount(); i++){
            anOrder = (Order)orderList.atOffset(i);
            totalSemimanufacturedRevenues += anOrder.getOrderEvaluation();
        }

        benefit = totalRevenues + dailyStoredComponentRevenues +
	    totalSemimanufacturedRevenues - totalCosts - 
	    totalInventoryFinCosts;

        return benefit;
    }
}


